# ------------------------------------------------------------------------------
# Unit tests for the clorm ORM interface
# ------------------------------------------------------------------------------

import unittest

from clorm.util.oset import OrderedSet

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

__all__ = [
    "OrderedSetTestCase",
]


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------


class OrderedSetTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    # --------------------------------------------------------------------------
    # --------------------------------------------------------------------------
    def test_basic_creation_iteration(self):

        # Creating an OrderedSet with a list and will additional items
        inlist1 = ["a", "b", "c", "d"]
        oset = OrderedSet(inlist1)
        self.assertEqual(list(oset), inlist1)

        oset = OrderedSet()
        for e in inlist1:
            oset.add(e)
        self.assertEqual(list(oset), inlist1)

        oset = OrderedSet(inlist1)
        inlist2 = [1, 2, 3]
        for e in inlist2:
            oset.add(e)
        self.assertEqual(list(oset), inlist1 + inlist2)

    def test_basic_ops(self):

        # Test pop
        inlist = ["a", "b", "c", "d"]
        oset = OrderedSet(inlist)
        oset.pop()
        self.assertEqual(list(oset), ["a", "b", "c"])
        oset.pop(last=False)
        self.assertEqual(list(oset), ["b", "c"])

        # Test pop for empty set
        oset = OrderedSet()
        with self.assertRaises(KeyError) as ctx:
            oset.pop()

        # Test clear
        inlist = ["a", "b", "c", "d"]
        oset = OrderedSet(inlist)
        oset.clear()
        self.assertEqual(list(oset), [])

        # Test remove and discard
        oset = OrderedSet(["a", "b", "c", "d"])
        oset.remove("b")
        self.assertEqual(list(oset), ["a", "c", "d"])

        # removing a non-existent item throws an exception
        with self.assertRaises(KeyError) as ctx:
            oset.remove("e")

        # discarding a non-existent item does nothing
        oset.discard("e")
        self.assertEqual(list(oset), ["a", "c", "d"])

        oset.discard("c")
        self.assertEqual(list(oset), ["a", "d"])

    def test_copy_equality(self):
        inlist = ["a", "b", "c", "d"]
        oset1 = OrderedSet(inlist)
        oset2 = oset1.copy()
        self.assertEqual(oset1, oset2)
        self.assertEqual(list(oset1), list(oset2))
        oset1.pop()
        self.assertFalse(list(oset1) == list(oset2))

        # Test equality
        inlist = ["a", "b", "c", "d"]
        oset1 = OrderedSet(inlist)
        inlist.reverse()
        oset2 = OrderedSet(inlist)
        oset3 = OrderedSet(["a", "b", "c"])

        # OrderedSets with the same elements but different ordering are not
        # equal
        self.assertFalse(oset1 == oset2)
        self.assertFalse(oset1 == oset3)

        # But an OrderedSet is equal to a normal set with the same elements and
        # different ordering.
        set1 = set(inlist)
        self.assertEqual(oset1, set1)
        self.assertFalse(oset3 == set1)

        # Test that the new isequal() function ignores the ordering
        self.assertFalse(oset1 == oset2)
        self.assertTrue(oset1.isequal(oset2))
        self.assertFalse(oset1.isequal(oset3))

    def test_bool_len_contains(self):
        inlist = ["a", "b", "c", "d"]
        oset = OrderedSet(inlist)
        self.assertEqual(len(oset), len(inlist))
        self.assertTrue("a" in oset)
        self.assertTrue("b" in oset)
        self.assertTrue("c" in oset)
        self.assertTrue("d" in oset)
        self.assertFalse("e" in oset)

        self.assertTrue(oset)
        oset.clear()
        self.assertFalse(oset)

    def test_bool_ops(self):
        set1 = set([1, 2, 3])
        oset1 = OrderedSet([1, 2, 3])

        # certain comparions cause a TypeError
        with self.assertRaises(TypeError) as ctx:
            set1 <= [1, 2, 3]
        with self.assertRaises(TypeError) as ctx:
            oset1 <= [1, 2, 3]
        with self.assertRaises(TypeError) as ctx:
            set1 < [1, 2, 3]
        with self.assertRaises(TypeError) as ctx:
            oset1 < [1, 2, 3]
        with self.assertRaises(TypeError) as ctx:
            set1 >= [1, 2, 3]
        with self.assertRaises(TypeError) as ctx:
            oset1 >= [1, 2, 3]
        with self.assertRaises(TypeError) as ctx:
            set1 > [1, 2, 3]
        with self.assertRaises(TypeError) as ctx:
            oset1 > [1, 2, 3]

        # Not sure why this doesn't raise a TypeError for set but I want the
        # behaviour to be consistent with OrderedSet
        self.assertFalse(set1 == [1, 2, 3])
        self.assertFalse(oset1 == [1, 2, 3])
        self.assertTrue(set1 != [1, 2, 3])
        self.assertTrue(oset1 != [1, 2, 3])

        self.assertTrue(set1.issubset([1, 2, 3]))
        self.assertTrue(oset1.issubset([1, 2, 3]))
        self.assertTrue(set1 <= set([1, 2, 3]))
        self.assertTrue(oset1 <= set([1, 2, 3]))
        self.assertTrue(set1 == set([1, 2, 3]))
        self.assertTrue(oset1 == set([1, 2, 3]))

        #        self.assertFalse(oset1 < [1,2,3])

        self.assertTrue(set1.issubset([1, 2, 3, 4]))
        self.assertTrue(oset1.issubset([1, 2, 3, 4]))
        self.assertTrue(set1 <= set([1, 2, 3, 4]))
        self.assertTrue(set1 <= OrderedSet([1, 2, 3, 4]))
        self.assertTrue(oset1 <= set([1, 2, 3, 4]))
        self.assertTrue(oset1 <= OrderedSet([1, 2, 3, 4]))

        self.assertTrue(set1 < set([1, 2, 3, 4]))
        self.assertTrue(set1 < OrderedSet([1, 2, 3, 4]))
        self.assertTrue(oset1 < set([1, 2, 3, 4]))
        self.assertTrue(oset1 < OrderedSet([1, 2, 3, 4]))

        self.assertFalse(set1.issubset([1, 2]))
        self.assertFalse(oset1.issubset([1, 2]))
        self.assertFalse(set1 <= set([1, 2]))
        self.assertFalse(oset1 <= set([1, 2]))
        self.assertFalse(set1 < set([1, 2]))
        self.assertFalse(oset1 < set([1, 2]))

        self.assertFalse(set1.issubset([2, 4]))
        self.assertFalse(oset1.issubset([2, 4]))
        self.assertFalse(set1 <= set([2, 4]))
        self.assertFalse(oset1 <= set([2, 4]))

        self.assertTrue(set1.issuperset([1, 2]))
        self.assertTrue(oset1.issuperset([1, 2]))
        self.assertTrue(set1 >= set([1, 2]))
        self.assertTrue(set1 >= OrderedSet([1, 2]))
        self.assertTrue(oset1 >= set([1, 2]))
        self.assertTrue(oset1 >= OrderedSet([1, 2]))

        self.assertTrue(set1.issuperset([1, 2, 3]))
        self.assertTrue(oset1.issuperset([1, 2, 3]))
        self.assertTrue(set1 >= set([1, 2, 3]))
        self.assertTrue(set1 >= OrderedSet([1, 2, 3]))
        self.assertTrue(oset1 >= set([1, 2, 3]))
        self.assertTrue(oset1 >= OrderedSet([1, 2, 3]))
        self.assertFalse(oset1.issuperset([1, 2, 3, 4]))
        self.assertFalse(set1 >= set([1, 2, 3, 4]))
        self.assertFalse(oset1 >= set([1, 2, 3, 4]))

        self.assertTrue(set1 > set([1, 2]))
        self.assertTrue(set1 > OrderedSet([1, 2]))
        self.assertTrue(oset1 > set([1, 2]))
        self.assertTrue(oset1 > OrderedSet([1, 2]))

        self.assertFalse(set1 > set([1, 2, 3]))
        self.assertFalse(oset1 > set([1, 2, 3]))
        self.assertFalse(oset1 > OrderedSet([1, 2, 3]))
        self.assertFalse(set1 > set([1, 2, 3, 4]))
        self.assertFalse(oset1 > set([1, 2, 3, 4]))

        self.assertFalse(set1.isdisjoint([1, 2, 3]))
        self.assertFalse(oset1.isdisjoint([1, 2, 3]))
        self.assertFalse(set1.isdisjoint([1, 2, 3, 4]))
        self.assertFalse(oset1.isdisjoint([1, 2, 3, 4]))
        self.assertFalse(set1.isdisjoint([1, 2]))
        self.assertFalse(oset1.isdisjoint([1, 2]))
        self.assertFalse(set1.isdisjoint([1, 2, 4]))
        self.assertFalse(oset1.isdisjoint([1, 2, 4]))

        self.assertTrue(oset1.isdisjoint([]))
        self.assertTrue(oset1.isdisjoint([4]))
        self.assertTrue(oset1.isdisjoint([4, 5]))

    def test_str_repr(self):
        set1 = set([])
        set2 = set([1])
        set3 = set([1, 2])
        set4 = set(["rar"])
        oset1 = OrderedSet([])
        oset2 = OrderedSet([1])
        oset3 = OrderedSet([1, 2])
        oset4 = OrderedSet(["rar"])

        self.assertEqual(str(set1), str(oset1))
        self.assertEqual(str(set2), str(oset2))
        self.assertEqual(str(set3), str(oset3))
        self.assertEqual(str(set4), str(oset4))

        self.assertEqual(repr(oset1), "OrderedSet()")
        self.assertEqual(repr(set2), repr(oset2))
        self.assertEqual(repr(set3), repr(oset3))
        self.assertEqual(repr(set4), repr(oset4))

    def test_set_union_intersection_functions(self):
        set1 = set([1, 2])
        set2 = set([2, 3])
        set3 = set([3, 4])
        oset1 = OrderedSet([1, 2])
        oset2 = OrderedSet([2, 3])
        oset3 = OrderedSet([3, 4])

        # Union/intersection/differeence of nothing returns a copy
        self.assertEqual(oset1.union(), set1.union())
        self.assertEqual(oset1.intersection(), set1.intersection())
        self.assertEqual(oset1.difference(), set1.difference())

        # Make sure the behaviour for OrderedSet operating on non-sets is the
        # same as for normal set.
        tmp1 = set1.union([2, 3])
        otmp1 = oset1.union([2, 3])
        self.assertEqual(otmp1, tmp1)
        self.assertEqual(str(otmp1), str(tmp1))

        # The union function
        self.assertEqual(oset1.union(oset2), set1.union(set2))
        self.assertEqual(oset1.union(oset3), set1.union(set3))
        self.assertEqual(oset1.union(oset2, oset3), set1.union(set2, set3))

        # Union of OrderedSets preserves the order
        self.assertEqual(oset1.union(oset2, oset3), OrderedSet([1, 2, 3, 4]))

        # The intersection function
        self.assertEqual(oset1.intersection(oset2), set1.intersection(set2))
        self.assertEqual(oset1.intersection(oset3), set1.intersection(set3))
        self.assertEqual(oset1.intersection(oset2, oset3), set1.intersection(set2, set3))

        # Intersection of OrderedSets preserves the order
        otmp1 = OrderedSet([1, 2, 3, 4, 5, 6])
        otmp2 = otmp1.intersection(OrderedSet([2, 3, 4, 6]), OrderedSet([1, 2, 4, 5, 6]))
        self.assertEqual(otmp2, OrderedSet([2, 4, 6]))

        # The difference function
        self.assertEqual(oset1.difference(oset2), set1.difference(set2))
        self.assertEqual(oset1.difference(oset3), set1.difference(set3))
        self.assertEqual(oset1.difference(oset2, oset3), set1.difference(set2, set3))

        # Difference of OrderedSets preserves the order
        otmp1 = OrderedSet([1, 2, 3, 4, 5])
        otmp2 = otmp1.difference(OrderedSet([1, 4]), OrderedSet([2, 6]))
        self.assertEqual(otmp2, OrderedSet([3, 5]))

        # The symmetric_difference function
        self.assertEqual(oset1.symmetric_difference(oset2), set1.symmetric_difference(set2))
        self.assertEqual(oset1.symmetric_difference(oset3), set1.symmetric_difference(set3))

        # Symmetric_difference of OrderedSets preserves the order
        otmp1 = OrderedSet([1, 2, 4, 3])
        otmp2 = otmp1.symmetric_difference(OrderedSet([3, 4, 6, 5]))
        self.assertEqual(otmp2, OrderedSet([1, 2, 6, 5]))

    def test_update_set_union_intersection_functions(self):
        set1 = set([1, 2])
        set2 = set([2, 3])
        set3 = set([3, 4])
        oset1 = OrderedSet([1, 2])
        oset2 = OrderedSet([2, 3])
        oset3 = OrderedSet([3, 4])

        # Union/intersection/difference _update of nothing returns itself
        self.assertEqual(oset1.update(), set1.update())
        self.assertEqual(oset1.intersection_update(), set1.intersection_update())
        self.assertEqual(oset1.difference_update(), set1.difference_update())

        # The update function
        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.update(oset2)
        tmp1.update(set2)
        self.assertEqual(otmp1, tmp1)

        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.update(oset3)
        tmp1.update(set3)
        self.assertEqual(otmp1, tmp1)

        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.update(oset2, oset3)
        tmp1.update(set2, set3)
        self.assertEqual(otmp1, tmp1)

        # Updated of OrderedSets preserves the order
        otmp1 = OrderedSet(oset1)
        otmp1.update(oset2, oset3)
        self.assertEqual(otmp1, OrderedSet([1, 2, 3, 4]))

        # The intersection_update function
        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.intersection_update(oset2)
        tmp1.intersection_update(set2)
        self.assertEqual(otmp1, tmp1)

        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.intersection_update(oset3)
        tmp1.intersection_update(set3)
        self.assertEqual(otmp1, tmp1)

        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.intersection_update(oset2, oset3)
        tmp1.intersection_update(set2, set3)
        self.assertEqual(otmp1, tmp1)

        # Intersection_update of OrderedSets preserves the order
        otmp1 = OrderedSet([1, 2, 3, 4, 5, 6])
        otmp1.intersection_update(OrderedSet([2, 3, 4, 6]), OrderedSet([1, 2, 4, 5, 6]))
        self.assertEqual(otmp1, OrderedSet([2, 4, 6]))

        # The difference_update function
        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.difference_update(oset2)
        tmp1.difference_update(set2)
        self.assertEqual(otmp1, tmp1)

        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.difference_update(oset3)
        tmp1.difference_update(set3)
        self.assertEqual(otmp1, tmp1)

        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.difference_update(oset2, oset3)
        tmp1.difference_update(set2, set3)
        self.assertEqual(otmp1, tmp1)

        # Difference_update of OrderedSets preserves the order
        otmp1 = OrderedSet([1, 2, 3, 4, 5])
        otmp1.difference_update(OrderedSet([1, 4]), OrderedSet([2, 6]))
        self.assertEqual(otmp1, OrderedSet([3, 5]))

        # The symmetric_difference_update function
        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.symmetric_difference_update(oset2)
        tmp1.symmetric_difference_update(set2)
        self.assertEqual(otmp1, tmp1)

        otmp1 = OrderedSet(oset1)
        tmp1 = set(set1)
        otmp1.symmetric_difference_update(oset3)
        tmp1.symmetric_difference_update(set3)
        self.assertEqual(otmp1, tmp1)

        # Symmetric_difference_update of OrderedSets preserves the order
        otmp1 = OrderedSet([1, 2, 4, 3])
        otmp1.symmetric_difference_update(OrderedSet([3, 4, 6, 5]))
        self.assertEqual(otmp1, OrderedSet([1, 2, 6, 5]))

    def test_set_union_intersection_operators(self):
        set1 = set([1, 2])
        set2 = set([2, 3])
        set3 = set([3, 4])
        oset1 = OrderedSet([1, 2])
        oset2 = OrderedSet([2, 3])
        oset3 = OrderedSet([3, 4])

        self.assertEqual(oset1 | oset2, set1 | set2)
        self.assertEqual(oset1 | set2, set1 | set2)

        # I don't think I can do anything about `set | OrderedSet` being not supported
        #        self.assertEqual(set1|oset2, set1|set2)

        self.assertEqual(oset1 & oset2, set1 & set2)
        self.assertEqual(oset1 & set2, set1 & set2)

        self.assertEqual(oset1 - oset2, set1 - set2)
        self.assertEqual(oset1 - set2, set1 - set2)

        self.assertEqual(oset1 ^ oset2, set1 ^ set2)
        self.assertEqual(oset1 ^ set2, set1 ^ set2)


# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



# ------------------------------------------------------------------------------
# Unit tests for the clorm ORM interface
# ------------------------------------------------------------------------------

import json
import unittest

import clingo

import clorm.json as cjson
from clorm import ComplexTerm, FactBase, IntegerField, Predicate, StringField

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

__all__ = ["JSONSymbolTestCase", "JSONPredicateTestCase"]

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------


class JSONSymbolTestCase(unittest.TestCase):
    def setUp(self):
        self.inf = clingo.Infimum
        self.sup = clingo.Supremum
        self.s1 = clingo.String("aaaa")
        self.s2 = clingo.String("bbbb")
        self.n1 = clingo.Number(24)
        self.n2 = clingo.Number(60)
        self.f1 = clingo.Function("", [self.s1, self.s2])
        self.f2 = clingo.Function("func1", [self.n1])
        self.f3 = clingo.Function("func2", [self.n2, self.f1])
        self.f4 = clingo.Function("func1", [], False)
        self.alls = [
            self.inf,
            self.sup,
            self.s1,
            self.s2,
            self.n1,
            self.n2,
            self.f1,
            self.f2,
            self.f3,
            self.f4,
        ]

        self.str_inf = '{"clingo.SymbolType": "Infimum"}'
        self.str_sup = '{"clingo.SymbolType": "Supremum"}'
        self.str_s1 = '{"clingo.SymbolType": "String", "string": "aaaa"}'
        self.str_s2 = '{"clingo.SymbolType": "String", "string": "bbbb"}'
        self.str_n1 = '{"clingo.SymbolType": "Number", "number": 24}'
        self.str_n2 = '{"clingo.SymbolType": "Number", "number": 60}'
        self.str_f1 = (
            '{"clingo.SymbolType": "Function", "name": "", "arguments": ['
            + self.str_s1
            + ", "
            + self.str_s2
            + '], "positive": true}'
        )
        self.str_f2 = (
            '{"clingo.SymbolType": "Function", "name": "func1", "arguments": ['
            + self.str_n1
            + '], "positive": true}'
        )
        self.str_f3 = (
            '{"clingo.SymbolType": "Function", "name": "func2", "arguments": ['
            + self.str_n2
            + ", "
            + self.str_f1
            + '], "positive": true}'
        )
        self.str_f4 = (
            '{"clingo.SymbolType": "Function", "name": "func1", "arguments": []'
            + ', "positive": false}'
        )
        self.str_alls = (
            "["
            + self.str_inf
            + ", "
            + self.str_sup
            + ", "
            + self.str_s1
            + ", "
            + self.str_s2
            + ", "
            + self.str_n1
            + ", "
            + self.str_n2
            + ", "
            + self.str_f1
            + ", "
            + self.str_f2
            + ", "
            + self.str_f3
            + ", "
            + self.str_f4
            + "]"
        )

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_symbol_encoder(self):
        inf = self.inf
        sup = self.sup
        s1 = self.s1
        s2 = self.s2
        n1 = self.n1
        n2 = self.n2
        f1 = self.f1
        f2 = self.f2
        f3 = self.f3
        alls = self.alls

        str_inf = self.str_inf
        str_sup = self.str_sup
        str_s1 = self.str_s1
        str_s2 = self.str_s2
        str_n1 = self.str_n1
        str_n2 = self.str_n2
        str_f1 = self.str_f1
        str_f2 = self.str_f2
        str_f3 = self.str_f3
        str_alls = self.str_alls

        encoder = cjson.symbol_encoder
        js_inf = json.dumps(inf, default=encoder)
        js_sup = json.dumps(sup, default=encoder)
        js_s1 = json.dumps(s1, default=encoder)
        js_s2 = json.dumps(s2, default=encoder)
        js_n1 = json.dumps(n1, default=encoder)
        js_n2 = json.dumps(n2, default=encoder)
        js_f1 = json.dumps(f1, default=encoder)
        js_f2 = json.dumps(f2, default=encoder)
        js_f3 = json.dumps(f3, default=encoder)
        js_alls = json.dumps(alls, default=encoder)

        self.assertEqual(js_inf, str_inf)
        self.assertEqual(js_sup, str_sup)
        self.assertEqual(js_s1, str_s1)
        self.assertEqual(js_s2, str_s2)
        self.assertEqual(js_n1, str_n1)
        self.assertEqual(js_n2, str_n2)

        self.assertEqual(js_f1, str_f1)
        self.assertEqual(js_f2, str_f2)
        self.assertEqual(js_f3, str_f3)
        self.assertEqual(js_alls, str_alls)

    def test_symbol_decoder(self):
        inf = self.inf
        sup = self.sup
        s1 = self.s1
        s2 = self.s2
        n1 = self.n1
        n2 = self.n2
        f1 = self.f1
        f2 = self.f2
        f3 = self.f3
        alls = self.alls

        str_inf = self.str_inf
        str_sup = self.str_sup
        str_s1 = self.str_s1
        str_s2 = self.str_s2
        str_n1 = self.str_n1
        str_n2 = self.str_n2
        str_f1 = self.str_f1
        str_f2 = self.str_f2
        str_f3 = self.str_f3
        str_alls = self.str_alls

        decoder = cjson.symbol_decoder
        self.assertEqual(json.loads(str_inf, object_hook=decoder), inf)
        self.assertEqual(json.loads(str_sup, object_hook=decoder), sup)
        self.assertEqual(json.loads(str_s1, object_hook=decoder), s1)
        self.assertEqual(json.loads(str_s2, object_hook=decoder), s2)
        self.assertEqual(json.loads(str_n1, object_hook=decoder), n1)
        self.assertEqual(json.loads(str_n2, object_hook=decoder), n2)
        self.assertEqual(json.loads(str_f1, object_hook=decoder), f1)
        self.assertEqual(json.loads(str_f2, object_hook=decoder), f2)
        self.assertEqual(json.loads(str_f3, object_hook=decoder), f3)
        self.assertEqual(json.loads(str_alls, object_hook=decoder), alls)


# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------


class JSONPredicateTestCase(unittest.TestCase):
    def setUp(self):
        class Fun(ComplexTerm):
            aint = IntegerField()
            astr = StringField()

        class Tup(ComplexTerm):
            aint = IntegerField()
            astr = StringField()

            class Meta:
                is_tuple = True

        class Afact(Predicate):
            aint = IntegerField()
            afun = Fun.Field()

        class Bfact(Predicate):
            astr = StringField()
            atup = Tup.Field()

        class Cfact(ComplexTerm):
            aint = IntegerField()
            astr = StringField()

        afact1 = Afact(aint=10, afun=Fun(aint=1, astr="a"))
        afact2 = Afact(aint=20, afun=Fun(aint=2, astr="b"))
        bfact1 = Bfact(astr="aa", atup=Tup(aint=1, astr="a"))
        bfact2 = Bfact(astr="bb", atup=Tup(aint=2, astr="b"))
        self.allf = [afact1, bfact1, afact2, bfact2]

        self.Fun = Fun
        self.Tup = Tup
        self.Afact = Afact
        self.Bfact = Bfact
        self.Cfact = Cfact

        self.n1 = clingo.Number(60)
        self.s1 = clingo.String("aaaa")
        self.f1 = clingo.Function("", [self.n1, self.s1])
        self.p1 = Cfact(aint=60, astr="aaaa")
        self.fb1 = FactBase(facts=[self.p1])

        self.str_n1 = '{"clingo.SymbolType": "Number", "number": 60}'
        self.str_s1 = '{"clingo.SymbolType": "String", "string": "aaaa"}'
        self.str_f1 = (
            '{"clingo.SymbolType": "Function", "name": "cfact", '
            + '"arguments": ['
            + self.str_n1
            + ", "
            + self.str_s1
            + "]"
            + ', "positive": true}'
        )
        self.str_p1 = '{"clorm.Predicate": "Cfact", "raw": ' + self.str_f1 + "}"
        self.str_fb1 = '{"clorm.FactBase": [], "facts": [' + self.str_p1 + "]}"

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_predicate_coder(self):
        pc1 = cjson.FactBaseCoder()
        Afact = pc1.register(self.Afact)
        Bfact = pc1.register(self.Bfact)
        Cfact = pc1.register(self.Cfact)
        allf = self.allf
        p1 = self.p1
        str_p1 = self.str_p1

        pc2 = cjson.FactBaseCoder([Afact, Bfact, Cfact])
        json_str1 = pc1.dumps(allf)
        json_str2 = pc2.dumps(allf)
        result1 = pc1.loads(json_str2)
        result2 = pc2.loads(json_str1)
        self.assertEqual(allf, result1)
        self.assertEqual(allf, result2)

        json_p1 = pc2.dumps(p1)
        self.assertEqual(json_p1, str_p1)

    # --------------------------------------------------------------------------
    #
    # --------------------------------------------------------------------------
    def test_factbase_coder(self):
        pc = cjson.FactBaseCoder()
        allf = self.allf
        fb1 = self.fb1
        str_fb1 = self.str_fb1
        Afact = pc.register(self.Afact)
        Bfact = pc.register(self.Bfact)
        Cfact = pc.register(self.Cfact)

        json_fb1 = pc.dumps(fb1)
        self.assertEqual(json_fb1, str_fb1)

        fb_in = FactBase(facts=allf, indexes=[Afact.aint, Bfact.astr])
        json_str = pc.dumps(fb_in, indent=4, sort_keys=True)
        fb_out = pc.loads(json_str)
        self.assertEqual(fb_in.indexes, fb_out.indexes)
        self.assertEqual(set(fb_in), set(fb_out))
        self.assertEqual(fb_in, fb_out)


# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")



#!/usr/bin/env python

import os
import re
import sys

from setuptools import setup


# Utility functions so that we can populate the package description and the
# version number automatically.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def find_version(fname):
    version_file = read(fname)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="Clorm",
    version = '1.6.1',
    author="David Rajaratnam",
    author_email="daver@gemarex.com.au",
    description="Clingo ORM (CLORM) provides a ORM interface for interacting with the Clingo Answer Set Programming (ASP) solver",
    license="MIT",
    url="https://github.com/potassco/clorm",
    packages=["clorm", "clorm.orm", "clorm.util", "clorm.lib"],
    package_data={"clorm": ["py.typed"]},
    zip_safe=False,  # https://mypy.readthedocs.io/en/latest/installed_packages.html
    install_requires=["clingo"],
    long_description=read("README.rst"),
)



"""
Extend clingo's Control, Model and SolveHandle classes to better use with clorm
"""

# TODO: For clorm v2.0 the raise_on_empty parameter in Model.facts() should be
#       moved to the second parameter position.

from __future__ import annotations

import functools
import itertools
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

import clingo as oclingo

from .orm import FactBase, Predicate, Symbol, SymbolPredicateUnifier, control_add_facts
from .util.wrapper import init_wrapper, make_class_wrapper

__all__ = ["ClormControl", "ClormModel", "ClormSolveHandle", "_expand_assumptions"]

OModel = oclingo.Model
OSolveHandle = oclingo.SolveHandle
OControl = oclingo.Control

# ------------------------------------------------------------------------------
# Helper function to smartly build a unifier if only a list of predicates have
# been provided.
# ------------------------------------------------------------------------------


_Unifier = Union[List[Type[Predicate]], SymbolPredicateUnifier]


def _build_unifier(unifier: Optional[_Unifier]) -> Optional[SymbolPredicateUnifier]:
    if unifier is None:
        return None
    if isinstance(unifier, SymbolPredicateUnifier):
        return unifier
    return SymbolPredicateUnifier(predicates=unifier)


# ------------------------------------------------------------------------------
# Helper function to test that an attribute exists and corresponds to a function.
# ------------------------------------------------------------------------------


def _check_is_func(obj: Any, name: str) -> None:
    if not callable(obj.__getattribute__(name)):
        raise AttributeError(
            ("Wrapped object of type '{}' does not have " "a function '{}()'").format(
                type(obj), name
            )
        )


# ------------------------------------------------------------------------------
# Wrap clingo.Model and override some functions
# ------------------------------------------------------------------------------


# class Model(OModel, metaclass=WrapperMetaClass):
class ModelOverride(object):
    """Provides access to a model during a solve call.

    Objects mustn't be created manually. Instead they are returned by
    ``clorm.clingo.Control.solve`` callbacks.

    Behaves like ``clingo.Model`` but offers better integration with clorm facts
    and fact bases.

    """

    if TYPE_CHECKING:
        _wrapped: OModel  # will be set through init_wrapper

    def __init__(self, model: OModel, unifier: Optional[_Unifier] = None) -> None:
        self._unifier = _build_unifier(unifier)
        _check_is_func(model, "symbols")
        init_wrapper(self, wrapped_=model)

    # ------------------------------------------------------------------------------
    # Return the underlying model object
    # ------------------------------------------------------------------------------
    @property
    def model_(self) -> OModel:
        """Returns the underlying clingo.Model object."""
        return self._wrapped

    # ------------------------------------------------------------------------------
    # A new function to return a list of facts - similar to symbols
    # ------------------------------------------------------------------------------

    def facts(self, *args: Any, **kwargs: Any) -> FactBase:
        """Returns a FactBase containing the facts in the model that unify with the
        SymbolPredicateUnifier.

        This function provides a wrapper around the ``clingo.Model.symbols``
        functions, but instead of returning a list of symbols it returns a
        FactBase containing the facts represented as ``clorm.Predicate``
        sub-class instances.

        Args:
           unifier(list | SymbolPredicateUnifier): used to unify and instantiate
              FactBase (Default: passed via the constructor if specified in the
              `clorm.clingo.Control` object)
           atoms: select all atoms in the model (Default: False)
           terms: select all terms displayed with #show statements (Default: False)
           shown: select all atoms and terms (Default: False)
           theory: select atoms added with Model.extend() (Default: False)
           complement: return the complement of the answer set w.r.t. to the
                       atoms known to the grounder. (Default: False)
           raise_on_empty: raise a ValueError if the resulting FactBase is empty
                           (Default: False)

        """
        nargs = list(args)
        nkwargs = dict(kwargs)
        if len(nargs) >= 1 and "unifier" in nkwargs:
            raise TypeError("facts() got multiple values for argument 'unifier'")
        if len(nargs) >= 7 and "raise_on_empty" in nkwargs:
            raise TypeError("facts() got multiple values for argument 'raise_on_empty'")

        raise_on_empty = nkwargs.pop("raise_on_empty", False)
        if len(nargs) >= 7:
            raise_on_empty = nargs.pop(6)
        unifier = nkwargs.pop("unifier", None)
        if len(nargs) >= 1:
            unifier = nargs.pop(0)

        unifier_ = _build_unifier(unifier) if unifier is not None else self._unifier
        if unifier_ is None:
            msg = (
                "Missing a predicate unifier specification in function call "
                + "(no default was given at model instantiation)"
            )
            raise ValueError(msg)

        return unifier_.unify(
            symbols=self.model_.symbols(*nargs, **nkwargs),
            raise_on_empty=raise_on_empty,
            delayed_init=True,
        )

    # ------------------------------------------------------------------------------
    # Overide contains
    # ------------------------------------------------------------------------------

    def contains(self, fact: Union[Predicate, oclingo.Symbol]) -> bool:
        """Return whether the fact or symbol is contained in the model. Extends
        ``clingo.Model.contains`` to allow for clorm facts as well as a clingo
        symbols.

        """
        atom = fact.raw if isinstance(fact, Predicate) else fact
        return self.model_.contains(atom)


if TYPE_CHECKING:

    class ClormModel(ModelOverride, OModel):
        pass

else:
    ClormModel = make_class_wrapper(OModel, ModelOverride)

# ------------------------------------------------------------------------------
# Wrap clingo.SolveHandle and override some functions
# ------------------------------------------------------------------------------


# class SolveHandle(OSolveHandle, metaclass=WrapperMetaClass):
class SolveHandleOverride(object):
    """Handle for solve calls.

    Objects mustn't be created manually. Instead they are returned by
    ``clorm.clingo.Control.solve``.

    Behaves like ``clingo.SolveHandle`` but iterates over ``clorm.clingo.Model``
    objects.

    """

    if TYPE_CHECKING:
        _wrapped: OSolveHandle  # will be set through init_wrapper

    def __init__(self, handle: OSolveHandle, unifier: Optional[_Unifier] = None) -> None:
        init_wrapper(self, wrapped_=handle)
        self._unifier = _build_unifier(unifier)

    # ------------------------------------------------------------------------------
    # Return the underlying solvehandle object
    # ------------------------------------------------------------------------------
    @property
    def solvehandle_(self) -> OSolveHandle:
        """Access the underlying clingo.SolveHandle object."""
        return self._wrapped

    # ------------------------------------------------------------------------------
    # Overrides
    # ------------------------------------------------------------------------------

    def __iter__(self):
        for model in self.solvehandle_:
            yield ClormModel(model, unifier=self._unifier)

    def __enter__(self):
        self.solvehandle_.__enter__()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.solvehandle_.__exit__(exception_type, exception_value, traceback)
        return None


if TYPE_CHECKING:

    class ClormSolveHandle(SolveHandleOverride, OSolveHandle):
        pass

else:
    ClormSolveHandle = make_class_wrapper(OSolveHandle, SolveHandleOverride)


# ------------------------------------------------------------------------------
# Wrap clingo.Control and override some functions
# ------------------------------------------------------------------------------


# ------------------------------------------------------------------------------
# Helper functions to expand the assumptions list as part of a solve() call. The
# assumptions list is a list of argument-boolean pairs where the argument can be
# be a clingo symbol, or clorm predicate instance, or a collection of clingo
# symbols or clorm predicates. This needs to be expanded into a list of
# symbol-bool pairs.
# ------------------------------------------------------------------------------
def _expand_assumptions(
    assumptions: Iterable[
        Tuple[Union[Iterable[Union[Predicate, Symbol]], Predicate, Symbol], bool]
    ],
) -> List[Tuple[Symbol, bool]]:
    clingo_assump = []

    def _add_fact(fact: Union[Predicate, Symbol], bval: bool) -> None:
        raw = fact.raw if isinstance(fact, Predicate) else fact
        clingo_assump.append((raw, bool(bval)))

    try:
        for arg, bval in assumptions:
            if isinstance(arg, Predicate):
                _add_fact(arg, bval)
            elif isinstance(arg, Iterable):
                for f in arg:
                    _add_fact(cast(Union[Predicate, Symbol], f), bval)
            else:
                _add_fact(arg, bval)
    except (TypeError, ValueError) as e:
        raise TypeError(
            (
                "Invalid solve assumptions. Expecting list of arg-bool "
                "pairs (arg is a raw-symbol/predicate or a collection "
                "of raw-symbols/predicates). Got: {}"
            ).format(assumptions)
        )
    return clingo_assump


# ------------------------------------------------------------------------------
# Control class
# ------------------------------------------------------------------------------


# class Control(OControl, metaclass=WrapperMetaClass):
class ControlOverride(object):
    """Control object for the grounding/solving process.

    Behaves like ``clingo.Control`` but with modifications to deal with Clorm
    facts and fact bases.

    Adds an additional parameter ``unifier`` to specify how any generated clingo
    models will be unified with the clorm Predicate definitions. The unifier can
    be specified as a list of predicates or as a SymbolPredicateUnifier object.

    An existing ``clingo.Control`` object can be passed using the ``control_``
    parameter.

    """

    if TYPE_CHECKING:
        _wrapped: OControl  # will be set through init_wrapper

    @overload
    def __init__(
        self,
        arguments: Sequence[str] = [],
        logger: Optional[oclingo.Logger] = None,
        message_limit: int = 20,
        unifier: Optional[Union[List[Predicate], SymbolPredicateUnifier]] = None,
    ) -> None: ...

    @overload
    def __init__(self, control_: OControl) -> None: ...

    @overload
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._unifier = None
        if "unifier" in kwargs:
            self._unifier = _build_unifier(kwargs["unifier"])

        # Do we need to build a clingo.Control object or use an existing one. If
        # using existing one make sure the wrapped object has at least the
        # ground and solve functions.
        if len(args) == 0 and "control_" in kwargs:
            wrapped_ = kwargs["control_"]
            _check_is_func(wrapped_, "solve")
            _check_is_func(wrapped_, "ground")
            init_wrapper(self, wrapped_=wrapped_)
        else:
            kwargs2 = dict(kwargs)
            if "unifier" in kwargs2:
                del kwargs2["unifier"]
            init_wrapper(self, *args, **kwargs2)

    # ------------------------------------------------------------------------------
    # Return the underlying control object
    # ------------------------------------------------------------------------------
    @property
    def control_(self) -> OControl:
        """Returns the underlying clingo.Control object."""
        return self._wrapped

    # ------------------------------------------------------------------------------
    # Make the unifier a property with a getter and setter
    # ------------------------------------------------------------------------------
    @property
    def unifier(self) -> Optional[SymbolPredicateUnifier]:
        """Get/set the unifier.

        Unifier can be specified as a SymbolPredicateUnifier or a collection of
        Predicates. Always returns a SymbolPredicateUnifier (or None).

        """
        return self._unifier

    @unifier.setter
    def unifier(self, unifier: _Unifier) -> None:
        self._unifier = _build_unifier(unifier)

    # ------------------------------------------------------------------------------
    # A new function to add facts from a factbase or a list of facts
    # ------------------------------------------------------------------------------
    def add_facts(self, facts: Iterable[Union[Predicate, Symbol]]) -> None:
        """Add facts to the control object. Note: facts must be added before grounding.

           This function can take an arbitrary collection containing a mixture
           of ``clorm.Predicate`` and ``clingo.Symbol`` objects. A
           ``clorm.FactBase`` is also a valid collection but it can only contain
           ``clorm.Predicate`` instances.

        Args:
          facts: a collection of ``clorm.Predicate`` or ``clingo.Symbol`` objects

        """
        control_add_facts(self.control_, facts)

    # ------------------------------------------------------------------------------
    # Overide assign_external to deal with Predicate object and a Clingo Symbol
    # ------------------------------------------------------------------------------
    def assign_external(
        self, external: Iterable[Union[Predicate, oclingo.Symbol, int]], truth: Optional[bool]
    ) -> None:
        """Assign a truth value to an external fact (or collection of facts)

        A fact can be a raw clingo.Symbol object, a clorm.Predicate instance, or
        a program literal (an int). If the external is a collection then the
        truth value is assigned to all elements in the collection.

        This function extends ``clingo.Control.release_external``.

        """

        def _assign_fact(fact: Union[Predicate, oclingo.Symbol, int]) -> None:
            if isinstance(fact, Predicate):
                self.control_.assign_external(fact.raw, truth)
            else:
                self.control_.assign_external(fact, truth)

        if isinstance(external, Iterable):
            for f in external:
                _assign_fact(f)
        else:
            _assign_fact(external)

    # ------------------------------------------------------------------------------
    # Overide release_external to deal with Predicate object and a Clingo Symbol
    # ------------------------------------------------------------------------------
    def release_external(self, external: Iterable[Union[Predicate, oclingo.Symbol, int]]) -> None:
        """Release an external fact (or collection of facts)

        A fact can be a raw clingo.Symbol object, a clorm.Predicate instance, or
        a program literal (an int). If the external is a collection then the
        truth value is assigned to all elements in the collection.

        This function extends ``clingo.Control.release_external``.

        """

        def _release_fact(fact: Union[Predicate, oclingo.Symbol, int]) -> None:
            if isinstance(fact, Predicate):
                self.control_.release_external(fact.raw)
            else:
                self.control_.release_external(fact)

        if isinstance(external, Iterable):
            for f in external:
                _release_fact(f)
        else:
            _release_fact(external)

    # ---------------------------------------------------------------------------
    # Overide solve and if necessary replace on_model with a wrapper that
    # returns a clorm.Model object. Also because of the issue with using the
    # keyword "async" as a parameter in Python 3.7 (which forced clingo 5.3.1+
    # to switch to using "async_") we have to use a complicated way to work out
    # function parameters. At some point will drop support for older clingo and
    # can simplify this function.
    # ---------------------------------------------------------------------------
    def solve(self, *args: Any, **kwargs: Any) -> Union[ClormSolveHandle, oclingo.SolveResult]:
        """Run the clingo solver.

        This function extends ``clingo.Control.solve()`` in two ways:

        1) The ``assumptions`` argument is generalised so that in the list of
        argument-boolean pairs the argument can be be a clingo symbol, or clorm
        predicate instance, or a collection of clingo symbols or clorm
        predicates.

        2) It produces either a ``clorm.clingo.SolveHandle`` wrapper object or a
        ``clorm.clingo.Model`` wrapper objects as appropriate (depending on the
        ``yield_``, ``async_``, and ``on_model`` parameters).

        """

        # Build the list of valid arguments; using the correct "async" or
        # "async_" parameter based on the clingo version.  Note: "async" is a
        # keyword for Python 3.7+.
        async_keyword = "async_" if oclingo.__version__ > "5.3.1" else "async"

        posnargs = ["assumptions", "on_model"]

        if oclingo.__version__ >= "5.5.0":
            posnargs.append("on_unsat")
        posnargs.extend(["on_statistics", "on_finish"])
        if oclingo.__version__ >= "5.5.0":
            posnargs.append("on_core")
        if oclingo.__version__ >= "5.8.0":
            posnargs.append("on_last")
        posnargs.extend(["yield_", async_keyword])
        validargs = set(posnargs)

        # translate all positional arguments into keyword arguments.
        if len(args) > len(posnargs):
            raise TypeError(
                ("solve() takes {} positional arguments but {}" "were given").format(
                    len(posnargs), len(args)
                )
            )
        nkwargs = {posnargs[idx]: arg for idx, arg in enumerate(args)}

        for k, v in kwargs.items():
            if k not in validargs:
                raise TypeError(("solve() got an unexpected keyword " "argument '{}'").format(k))
            if k in nkwargs:
                raise TypeError(("solve() got multiple values for " "argument '{}'").format(k))
            nkwargs[k] = v

        # generate a new assumptions list if necesary
        if "assumptions" in nkwargs and nkwargs["assumptions"] is not None:
            nkwargs["assumptions"] = _expand_assumptions(nkwargs["assumptions"])

        # generate a new on_model function if necessary
        if "on_model" in nkwargs and nkwargs["on_model"] is not None:
            on_model = nkwargs["on_model"]

            @functools.wraps(on_model)
            def on_model_wrapper(model):
                return on_model(ClormModel(model, self.unifier))

            nkwargs["on_model"] = on_model_wrapper

        # generate a new on_last function if necessary
        if (
            oclingo.__version__ >= "5.8.0"
            and "on_last" in nkwargs
            and nkwargs["on_last"] is not None
        ):
            on_last = nkwargs["on_last"]

            @functools.wraps(on_last)
            def on_last_wrapper(model):
                return on_last(ClormModel(model, self.unifier))

        # Call the wrapped solve function and handle the return value
        # appropriately
        result = self.control_.solve(**nkwargs)
        if ("yield_" in nkwargs and nkwargs["yield_"]) or (
            async_keyword in nkwargs and nkwargs[async_keyword]
        ):
            return ClormSolveHandle(cast(OSolveHandle, result), unifier=self._unifier)
        else:
            return cast(oclingo.SolveResult, result)

    def __getattr__(self, attr):
        return getattr(self.control_, attr)


if TYPE_CHECKING:

    class ClormControl(ControlOverride, OControl):  # type: ignore
        pass

else:
    ClormControl = make_class_wrapper(OControl, ControlOverride)

# ------------------------------------------------------------------------------
# This is probably bad practice... Modify the original clingo docstrings so that
# when I generate the autodoc with clingo being mocked it installs a reference
# to the original clingo docs.
#
# UPDATE: I'm confused what the autodoc mocking is doing (or not doing). I'm
# sure I had it working but now seems to be failing. Adding more hacks :(
# ------------------------------------------------------------------------------

ClormControl.__doc__ += OControl.__doc__  # type: ignore
ClormControl.assign_external.__doc__ += OControl.assign_external.__doc__  # type: ignore
ClormControl.release_external.__doc__ += OControl.release_external.__doc__  # type: ignore
ClormControl.solve.__doc__ += OControl.solve.__doc__  # type: ignore
ClormModel.__doc__ += OModel.__doc__  # type: ignore
ClormModel.contains.__doc__ += OModel.contains.__doc__  # type: ignore
ClormSolveHandle.__doc__ += OSolveHandle.__doc__  # type: ignore

# ------------------------------------------------------------------------------
# main
# ------------------------------------------------------------------------------
if __name__ == "__main__":
    raise RuntimeError("Cannot run modules")
